#include "SD.h"
//#include "include.h"

/**
 *	Hardware Abstraction Layer initialization
 **/
void SDInit(void)
{
	//initalize pins, spi setup elsewhere (shared with CC1100)
	//CSn (output, high)
	SD_CSn_DIR&= ~SD_CSn_PIN; //0 for output (change for MSP430)
	SD_CSn_OUT|=SD_CSn_PIN; //1 for high
	
	//Card detect (input)
	SD_CD_DIR|=SD_CD_PIN; //1 for input (change for MSP430)

	//write protect (input)
//	SD_WP_DIR|=SD_WP_PIN; //1 for input (change for MSP430)
}


/**
 *	USES "CD" (card detect) switch on card holder....
 *
 *	@return	FALSE		card not detected
 *	@return TRUE		car detected
 **/
boolean detectCard(void)
{
	if(SD_CD_IN&SD_CD_PIN)
	{
		return TRUE;
	}
	return FALSE;
}

/**
 *	sends a byte on the SPI to the memory card
 *
 *	@param	inBuf		byte to be sent
 **/
void spi_put(unsigned char inBuf)
{
	SD_SPI_IF = 0;           	        // Clear flag
    SD_SSPBUF = inBuf;                    // Send address
    while(!SD_SPI_IF);			            // Wait for TX to finish
	SD_SPI_IF = 0;

}

/**
 *	reads a byte from the SPI to the memory card
 *
 *	@return	U1RXBUF		byte read
 **/
unsigned char spi_get(void)
{	
	char Dummy;

	SD_SPI_IF = 0;           	        // Clear flag
    SD_SSPBUF = 0xFF;                    // Send address
    while(!SD_SPI_IF);			            // Wait for TX to finish
	Dummy = SD_SSPBUF;
	SD_SPI_IF = 0;
	return Dummy;

}

/**
 *	sends the command to the the memory card with the 4 arguments and the CRC
 *
 *	@param command		the command
 *	@param argX			the arguments
 *	@param CRC			cyclic redundancy check
 **/
void sendCommand(unsigned char command, unsigned char argA,
                 unsigned char argB, unsigned char argC,
                 unsigned char argD, unsigned char CRC)
{	
	spi_put(command);
	spi_put(argA);
	spi_put(argB);
	spi_put(argC);
	spi_put(argD);
	spi_put(CRC);
}

/**
 *	waits for the memory card to response
 *
 *	@param expRes		expected response
 *
 *	@return TRUE		card responded within time limit
 *	@return FALSE		card did not response
 **/
boolean cardResponse(unsigned char expRes)
{
	unsigned char actRes;
	unsigned int count = 256;
	actRes = 0xFF;
	while ((actRes != expRes) && (count > 0))
	{
		actRes = spi_get();		
		count--;
	}
	if (count == 0)
		return FALSE; 
	return TRUE;
}

/**
 *	initializes the memory card by brign it out of idle state and 
 *	sets it up to use SPI mode
 *
 *	@return TRUE		card successfully initialized
 *	@return FALSE		card not initialized
 **/
byte memCardInit(void)
{
	unsigned int count;
	SD_CSn_OUT|=SD_CSn_PIN;	
	for (count = 0;count < 10; count++)
    {
		spi_put(0xFF);
    }
	SD_CSn_OUT&= ~SD_CSn_PIN;
	sendCommand(CMD0,0,0,0,0,0x95);	// initializes the MMC
	if(cardResponse(0x01))
	{				
		count = 255;
		do
		{
			sendCommand(CMD1,0,0,0,0,0xFF);	// brings card out of idle state
			count--;
		} while ((!cardResponse(0x00)) && (count > 0));
		SD_CSn_OUT|=SD_CSn_PIN;
		spi_put(0xFF);	
		if(count == 0)
			return 1;
	}
	else
	{
		SD_CSn_OUT|=SD_CSn_PIN;
		return 1;		
	}
	return 0;
}

/**
 *	Initialize the block length in the memory card to be 512 bytes
 *	
 *	return TRUE			block length sucessfully set
 *	return FALSE		block length not sucessfully set
 **/
boolean setBLockLength(void)
{
	SD_CSn_OUT&= ~SD_CSn_PIN;
	sendCommand(CMD16,0,0,2,0,0xFF);
	if(cardResponse(0x00))
	{
		spi_put(0xFF);
		SD_CSn_OUT|=SD_CSn_PIN;
		return TRUE;
	}
	spi_put(0xFF);
	SD_CSn_OUT|=SD_CSn_PIN;
	return FALSE;
}

/**
 *	checks to see if previous write operation is sucessful
 *	
 *	@return TRUE		if data is successfully written to disk
 *	@return FALSE		if data is not successfully written to disk
 **/
boolean checkWriteState(void)
{
	unsigned int count = 0;
	char inRes;
	while(count <= 64)
	{
		inRes = spi_get();
		if((inRes & 0x1F) == 0x05)
			break;
		else if((inRes & 0x1F) == 0x0B)
			return FALSE;
		else if((inRes & 0x1F) == 0x0D)
			return FALSE;
		count++;
	}
	count = 0;
	while((count < 64) && (spi_get() == 0x00));
	if(count < 64)
		return TRUE;
	else
		return FALSE;
}

/**
 *	read content of the sector indicated by the input and place it in the buffer
 *
 *	@param	sector		sector to be read
 *	@param	*buf		pointer to the buffer
 *
 *	@return	...			number of bytes read
 **/
//unsigned int readSector(unsigned long sector, unsigned char *buf)
uint32_t readSector(uint32_t sector, uint8_t *buf)
{
    unsigned int count = 0;
	//sector += sectorZero;
	sector *= SECTOR_SIZE;	
	SD_CSn_OUT&= ~SD_CSn_PIN;
	sendCommand(CMD17,(sector>>24)&0xFF,(sector>>16)&0xFF,(sector>>8)&0xFF,sector&0xFF,0xFF);
	if(cardResponse(0x00))
	{
		if(cardResponse(0xFE))
		{
			for(count = 0; count < SECTOR_SIZE; count++)
			{
				buf[count] = spi_get();
			}
			spi_get();
			spi_get();
		}
	}
	else
	{ 
		SD_CSn_OUT|=SD_CSn_PIN;
		spi_get();
		return 1; //error
	}
	SD_CSn_OUT|=SD_CSn_PIN;
	spi_get();
	return 0;
}


/**
 *	writes content of the buffer to the sector indicated by the input
 *
 *	@param	sector		sector to be written
 *	@param	*buf		pointer to the buffer
 *
 *	@return	...			number of bytes written
 **/
//unsigned int writeSector(unsigned long sector, unsigned char *buf)
uint32_t writeSector(uint32_t sector, uint8_t *buf)
{
    unsigned int count = 0;
//	BOOL stat;
//	sector += sectorZero;
	sector *= SECTOR_SIZE;	
	SD_CSn_OUT&= ~SD_CSn_PIN;
	sendCommand(CMD24,(sector>>24)&0xFF,(sector>>16)&0xFF,(sector>>8)&0xFF,sector&0xFF,0xFF);
//	sendCommand(CMD24,(sector>>16)&0xFF,(sector>>8)&0xFF,sector&0xFF,0,0xFF);
	if(cardResponse(0x00))
	{
		spi_put(0xFF);
		spi_put(0xFE);
		for(count = 0; count < SECTOR_SIZE; count++)
		{
			spi_put(buf[count]);
		}
		spi_get();
		spi_get();
		if(checkWriteState())
		{	//success
			SD_CSn_OUT|=SD_CSn_PIN;
			spi_get();
			return 0;
		}
		else
		{	//failed
			SD_CSn_OUT|=SD_CSn_PIN;
			spi_get();
			return 1;
		}
	}
	//total failure!
	SD_CSn_OUT|=SD_CSn_PIN;
	spi_get();
	return 1;
}

